/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package org.wso2.registry.app;

import org.apache.abdera.Abdera;
import org.apache.abdera.model.*;
import org.apache.abdera.protocol.Response;
import org.apache.abdera.protocol.client.AbderaClient;
import org.apache.abdera.protocol.client.ClientResponse;
import org.apache.axiom.om.util.Base64;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.Comment;
import org.wso2.registry.*;
import org.wso2.registry.i18n.Messages;

import javax.xml.namespace.QName;
import java.io.*;
import java.net.URL;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

public class RemoteRegistry implements Registry {

    private String collectionRoot = "http://localhost:8080/wso2registry/atom";
    private String entryRoot = "http://localhost:8080/wso2registry/feed";
    private Log log = LogFactory.getLog(RemoteRegistry.class);

    /**
     * To create a remote registry need to provide a URL of a remote registry and the URL
     * should be something like
     * http://localhost:8080/wso2registry
     * NOTE : Should not have "/" after the wso2registry part
     *
     * @param registryURL : URL to the registry
     * @throws RegistryException : If something went wrong
     */
    public RemoteRegistry(URL registryURL) throws RegistryException {
        String url = registryURL.toString();
        collectionRoot = url + "/atom";
        entryRoot = url + "/feed";
    }

    public Resource get(String path) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(collectionRoot + path);
        if (clientResponse.getType() == Response.ResponseType.SERVER_ERROR) {
            throw new RegistryException(Messages.getMessage("resource.not.found.warn", path));
        }
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();

        if (feed.getSubtitle() != null && feed.getSubtitle().equals("Deleted")) {
            throw new RegistryException(Messages.getMessage("resource.not.found"));
        }

        return createResourceFromFeed(feed);
    }

    private Resource createResourceFromFeed(Feed feed) {
        String value = feed.getAttributeValue("directory");
        Resource resource = new Resource();

        List attributes = feed.getAttributes();
        for (int i = 0; i < attributes.size(); i++) {
            QName attName = (QName) attributes.get(i);
            String key = attName.getLocalPart();
            if (!"directory".equals(key)) {
                resource.setProperty(key, feed.getAttributeValue(key));
            }
        }
        if (feed.getAuthor() != null) {
            resource.setAuthorUserName(feed.getAuthor().getName());
        }
        resource.setLastModified(feed.getUpdated());
        resource.setPath(feed.getTitle());
        resource.setDescription(feed.getText());
        String subtitle = feed.getSubtitle();
        if (subtitle != null && "Deleted".equals(subtitle)) {
            resource.setState(RegistryConstants.DELETED_STATE);
        }

        if (value != null && "true".equals(value)) {
            resource.setDirectory(true);
            List entries = feed.getEntries();
            if (entries != null) {
                String[] childNodes = new String[entries.size()];
                for (int i = 0; i < entries.size(); i++) {
                    Entry entery = (Entry) entries.get(i);
                    childNodes[i] = entery.getTitle();
                }
                resource.setContent(childNodes);
            }
        } else {
            List entries = feed.getEntries();
            Entry entery = (Entry) entries.get(0);
            String content1 = entery.getContent();

            if (content1 != null && !Content.Type.TEXT.equals(entery.getContentType())) {
                byte[] content = Base64.decode(content1);
                resource.setContent(content);
            } else {
                resource.setContent(content1);
            }
        }
        return resource;
    }

    public boolean resourceExists(String path) throws RegistryException {
        throw new UnsupportedOperationException(" Use get(path) instead");
    }

    public void put(String path, Resource resource) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setTitle(path);
        if (RegistryConstants.SQL_QUERY_MEDIA_TYPE.equals(resource.getMediaType())) {
            entry.setAttributeValue("mediaType", RegistryConstants.SQL_QUERY_MEDIA_TYPE);
        }
        entry.setSummary(resource.getDescription());
        entry.addAuthor(resource.getAuthorUserName());
        Properties properties = resource.getProperties();
        if (properties != null) {
            Iterator keys = properties.keySet().iterator();
            while (keys.hasNext()) {
                String key = (String) keys.next();
                entry.setAttributeValue(key, (String) properties.get(key));
            }
        }
        if (resource.isDirectory()) {
            entry.setAttributeValue("directory", "true");
        } else {
            entry.setAttributeValue("directory", "false");
        }

        if (!resource.isDirectory()) {
            Object content = resource.getContent();
            if (content != null && content instanceof byte[]) {
                ByteArrayInputStream in = new ByteArrayInputStream((byte[]) content);
                entry.setContent(in);
            } else if (content instanceof InputStream) {
                entry.setContent((InputStream) content);
            } else {
                entry.setContent((String) content);
            }
        }
        ClientResponse resp = abderaClient.post(collectionRoot + path, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.add", path));
        } else {
            log.error(Messages.getMessage("add.resource.fail", path));
            throw new RegistryException(Messages.getMessage("add.resource.fail", path));
        }
    }

    public void delete(String path) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse resp = abderaClient.delete(entryRoot + path);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.deleted", path));
        } else {
            log.error(Messages.getMessage("resource.delete.fail", path));
            throw new RegistryException(Messages.getMessage("resource.delete.fail", path));
        }
    }

    public String[] getVersions(String path) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(collectionRoot + path + "?" +
                RegistryConstants.PARAMETER_VERSION);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        String versionString = feed.getText();
        if (versionString != null) {
            return versionString.split(",");
        }
        return new String[0];
    }

    public void restoreVersion(String versionPath) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setTitle(versionPath);
        entry.setSummary(RegistryConstants.PARAMETER_RESTORE);
        entry.setContent(versionPath);
        entry.setAttributeValue("directory", "false");
        ClientResponse resp = abderaClient.put(entryRoot, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.restoreded", versionPath));
        } else {
            log.error(Messages.getMessage("resource.restore.fail", versionPath));
            throw new RegistryException(Messages.getMessage("resource.restore.fail", versionPath));
        }
    }

    public void applyTag(String resourcePath, String tag) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setTitle(resourcePath);
        entry.setSummary(RegistryConstants.PARAMETER_TAGS);
        entry.setContent(tag);
        entry.setAttributeValue("directory", "false");
        ClientResponse resp = abderaClient.put(entryRoot, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.tagged", resourcePath));
        } else {
            log.error(Messages.getMessage("resource.tag.fails", resourcePath));
            throw new RegistryException(Messages.getMessage("resource.tag.fails", resourcePath));
        }
    }

    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(collectionRoot + "?" +
                RegistryConstants.PARAMETER_TAG_PATHS + "=" + tag);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        List entries = feed.getEntries();
        TaggedResourcePath taggedResourcePaths[] = null;
        if (entries != null) {
            taggedResourcePaths = new TaggedResourcePath[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entery = (Entry) entries.get(i);
                TaggedResourcePath tagpath = new TaggedResourcePath();
                tagpath.setResourcePath(entery.getTitle());
                tagpath.setTagCount(Long.parseLong(entery.getContent()));
                taggedResourcePaths[i] = tagpath;
            }
        }
        return taggedResourcePaths;
    }

    public Tag[] getTags(String resourcePath) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(collectionRoot + resourcePath + "?" +
                RegistryConstants.PARAMETER_TAGS);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        List entries = feed.getEntries();
        Tag tags[] = null;
        if (entries != null) {
            tags = new Tag[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entery = (Entry) entries.get(i);
                Tag tag = new Tag();
                tag.setTagCount(Integer.parseInt(entery.getSummary()));
                tag.setTagName(entery.getTitle());
                tags[i] = tag;
            }
        }
        return tags;
    }

    public void removeTag(String path, String tag) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setTitle(path);
        entry.setContent(tag);
        entry.setSummary(RegistryConstants.PARAMETER_TAGS_REMOVED);
        ClientResponse resp = abderaClient.put(entryRoot, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("tag.removed", path));
        } else {
            log.error(Messages.getMessage("tag.removed.fails", path));
            throw new RegistryException(Messages.getMessage("tag.removed.fails", path));
        }
    }

    public void addComment(String resourcePath, Comment comment) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setTitle(resourcePath);
        entry.setSummary(RegistryConstants.PARAMETER_COMMENTS);
        entry.setUpdated(comment.getCommentedTime());
        entry.addAuthor(comment.getCommentedUser());
        entry.setContent(comment.getCommentText());
        ClientResponse resp = abderaClient.put(entryRoot, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.commented", resourcePath));
        } else {
            log.error(Messages.getMessage("resouce.comment.fails", resourcePath));
            throw new RegistryException(Messages.getMessage("resouce.comment.fails", resourcePath));
        }
    }

    public Comment[] getComments(String resourcePath) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(collectionRoot + resourcePath + "?" +
                RegistryConstants.PARAMETER_COMMENTS);
        Document introspection =
                clientResponse.getDocument();
        Element element = introspection.getRoot();
        Feed feed = (Feed) element;
        List entries = feed.getEntries();
        Comment comments[] = null;
        if (entries != null) {
            comments = new Comment[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entery = (Entry) entries.get(i);
                Comment comment = new Comment();
                comment.setCommentedTime(entery.getUpdated());
                comment.setCommentText(entery.getContent());
                comment.setCommentedUser(entery.getAuthor().getName());
                comments[i] = comment;
            }
        }
        return comments;
    }

    public void rateResource(String resourcePath, int rating) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setTitle(resourcePath);
        entry.setContent("" + rating);
        entry.setSummary(RegistryConstants.PARAMETER_REATINGS);
        ClientResponse resp = abderaClient.put(entryRoot, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.rated", resourcePath));
        } else {
            log.error(Messages.getMessage("resource.rate.fail", resourcePath));
            throw new RegistryException(Messages.getMessage("resource.rate.fail", resourcePath));
        }
    }

    public float getAverageRating(String resourcePath) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(collectionRoot + resourcePath + "?" +
                RegistryConstants.PARAMETER_AVARAGE_REATINGS);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        String floatvalue = feed.getSubtitle();
        return Float.parseFloat(floatvalue);
    }

    public int getRating(String path, String userName) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(collectionRoot + path + "/" + userName + "?" +
                RegistryConstants.PARAMETER_REATINGS);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        String intvalue = feed.getSubtitle();
        return Integer.parseInt(intvalue);
    }


    public Resource executeQuery(String path, Object[] parameters) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setTitle(path);
        entry.setSummary(RegistryConstants.PARAMETER_QUERY);
        entry.setContent(encodeParametesAsString(parameters));
        ClientResponse resp = abderaClient.put(entryRoot, entry);
        Document introspection =
                resp.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        return createResourceFromFeed(feed);
    }

    public static String encodeParametesAsString(Object[] parameters) {
        String value = "";
        for (int i = 0; i < parameters.length; i++) {
            Object parameter = parameters[i];
            if (i < parameters.length - 1) {
                value = value + parameter + "#";
            } else {
                value = value + parameter;
            }
        }
        return value;
    }

    public static String[] decodeParametesAsString(String value) {
        return value.split("\\#");
    }

    public LogEntry[] getLogs(String resourcePath,
                              int action,
                              String userName,
                              Date from,
                              Date to,
                              boolean recentFirst) throws RegistryException {

        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setEdited(to);
        entry.setUpdated(from);
        entry.setTitle("" + action);
        entry.setSummary(RegistryConstants.PARAMETER_LOGS);
        entry.setContent(resourcePath);
        entry.addAuthor(userName);
        ClientResponse resp = abderaClient.put(entryRoot, entry);
        Document introspection =
                resp.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        List entries = feed.getEntries();
        LogEntry logs[] = null;
        if (entries != null) {
            logs = new LogEntry[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entery = (Entry) entries.get(i);
                LogEntry logEntry = new LogEntry();
                logEntry.setDate(entery.getEdited());
                logEntry.setActionData(entery.getContent());
                logEntry.setUserName(entery.getAuthor().getName());
                logEntry.setAction(Integer.parseInt(entery.getSummary()));
                logEntry.setResourcePath(entery.getTitle());
                logs[i] = logEntry;
            }
        }
        return logs;
    }

    /**
     * This method can be used to export a local file system into a runing instanse of a registry.
     * Need to create a file object representing the local file and the need to tell where to add
     * the resource in the registry.
     * @param file : File representing local file system
     * @param path : Where to put the file
     * @throws RegistryException : if something went wrong
     */
    public void importToRegistry(File file, String path) throws RegistryException {
        try {
            if (file == null || path == null) {
                throw new RegistryException(Messages.getMessage("file.valaue.null"));
            }
            processImport(file, path);
        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
    }

    /**
     * This method can be used to export registry instance or node in a regsitry to a local file
     * system. When we call this method a file srcuture will be created to match the structure to
     * map the strcuture in the registry
     * @param toFile : File in the local file system
     * @param path : To which node export
     * @throws RegistryException : If something went wrong
     */
    public void exportFromRegistry(File toFile, String path) throws RegistryException {
        try {
            processExport(path, toFile);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RegistryException(e.getMessage());
        }
    }

    private void processImport(File file, String path) throws Exception {
        String filePart = file.getName();
        String resourcePath = path + "/" + filePart;
        if (file.isDirectory()) {
            File files[] = file.listFiles();
            if (files.length > 0) {
                for (int i = 0; i < files.length; i++) {
                    File childFile = files[i];
                    processImport(childFile, resourcePath);
                }
            } else {
                Resource resource = new Resource();
                resource.setPath(resourcePath);
                resource.setDirectory(true);
                put(resourcePath, resource);
            }
        } else {
            Resource resource = new Resource();
            resource.setContent(new FileInputStream(file));
            resource.setPath(resourcePath);
            resource.setDirectory(false);
            put(resourcePath, resource);
        }
    }

    private void processExport(String resourcePath,
                               File toFile) throws RegistryException {
        Resource resource = get(resourcePath);
        if (resource != null) {
            if (resource.isDirectory()) {
                File file = new File(toFile, resource.getPath());
                file.mkdirs();
                String childNodes[] = (String[]) resource.getContent();
                for (int i = 0; i < childNodes.length; i++) {
                    String childNode = childNodes[i];
                    processExport(childNode, toFile);
                }
            } else {
                File file = new File(toFile, resource.getPath());
                try {
                    file.createNewFile();
                    FileOutputStream out = new FileOutputStream(file);
                    out.write((byte[])resource.getContent());
                    out.flush();
                    out.close();
                } catch (IOException e) {
                    throw new RegistryException(Messages.getMessage(
                            "error.creating.file", resource.getPath()));
                }

            }
        }
    }
}
