/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package org.wso2.registry.config;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.wso2.registry.RegistryException;
import org.wso2.registry.i18n.Messages;

import javax.servlet.ServletConfig;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public class RegistryConfiguration {

    private DataBaseConfiguration deafultDataBaseConfiguration = null;
    private Map dbconfgigMap = new HashMap();

    /**
     * This method will read the regisrty.xml and using the data given in the file
     * resgistry configuratoon will be populated. And if the xmlLocation is null then it
     * will use the one in the class path.
     *
     * @param xmlLocation : Location of the
     * @throws org.wso2.registry.RegistryException
     *          : if something went wrong
     */
    public void  populateRegistryConfig(String xmlLocation, ServletConfig servletConfig) throws RegistryException {
        InputStream in;
        if (xmlLocation != null) {
            try {
                in = new FileInputStream(xmlLocation);
            } catch (FileNotFoundException e) {
                throw new RegistryException(e.getMessage());
            }
        } else {
            in = servletConfig.getServletContext().getResourceAsStream("/WEB-INF/registry.xml");
            if (in == null) {
                in = Thread.currentThread().getContextClassLoader().getResourceAsStream(
                        "org/wso2/servlet/registry.xml");
            }
            if (in == null) {
                throw new RegistryException(Messages.getMessage("inputstream.null",
                        "org/wso2/servlet/registry.xml"));
            }
        }
        try {
            StAXOMBuilder buillder = new StAXOMBuilder(in);
            OMElement configElement = buillder.getDocumentElement();
            if (configElement != null) {
                Iterator dbconfgs = configElement.getChildrenWithName(new QName("dbconfig"));
                while (dbconfgs.hasNext()) {
                    OMElement dbconfig = (OMElement) dbconfgs.next();
                    DataBaseConfiguration dataBaseConfiguration = new DataBaseConfiguration();

                    String dbName = dbconfig.getAttributeValue(new QName("name"));
                    if (dbName == null) {
                        throw new RegistryException(Messages.getMessage("dbname.null"));
                    }
                    dataBaseConfiguration.setConfigName(dbName);

                    OMElement dataSource = dbconfig.getFirstChildWithName(new QName("dataSource"));
                    if (dataSource != null) {
                        dataBaseConfiguration.setDataSourceName(dataSource.getText());

                    } else {

                        OMElement userName = dbconfig.getFirstChildWithName(new QName("userName"));
                        if (userName != null) {
                            dataBaseConfiguration.setUserName(userName.getText());
                        }

                        OMElement password = dbconfig.getFirstChildWithName(new QName("password"));
                        if (password != null) {
                            dataBaseConfiguration.setPassWord(password.getText());
                        }

                        OMElement url = dbconfig.getFirstChildWithName(new QName("url"));
                        if (url != null) {
                            dataBaseConfiguration.setDbUrl(url.getText());
                        }

                        OMElement driverName = dbconfig.getFirstChildWithName(new QName("driverName"));
                        if (driverName != null) {
                            dataBaseConfiguration.setDriverName(driverName.getText());
                        }
                    }

                    dbconfgigMap.put(dbName, dataBaseConfiguration);

                    String defaultConfig = dbconfig.getAttributeValue(new QName("default"));
                    if (defaultConfig != null && defaultConfig.equalsIgnoreCase("true")) {
                        setDeafultDataBaseConfiguration(dataBaseConfiguration);
                    }
                }
            }

        } catch (XMLStreamException e) {
            throw new RegistryException(e.getMessage());
        }
    }

    public void addDataBaseConfiguration(DataBaseConfiguration dataBaseConfiguration) {
        dbconfgigMap.put(dataBaseConfiguration.getConfigName(), dataBaseConfiguration);
    }

    public DataBaseConfiguration getDataBaseConfiguration(String configName) {
        return (DataBaseConfiguration) dbconfgigMap.get(configName);
    }

    public void removeDataBaseConfiguration(String configName) {
        dbconfgigMap.remove(configName);
    }

    public OMElement getRegistryConfigAsXML() {
        
        OMFactory factoty = OMAbstractFactory.getOMFactory();
        OMElement root = factoty.createOMElement("wso2regsitry", null);
        Iterator values = dbconfgigMap.values().iterator();
        
        while (values.hasNext()) {
            DataBaseConfiguration dataBaseConfiguration = (DataBaseConfiguration) values.next();
            OMElement config = factoty.createOMElement("dbconfig", null);
            OMElement url = factoty.createOMElement("url", null);
            url.setText(dataBaseConfiguration.getDbUrl());
            config.addChild(url);
            OMElement userName = factoty.createOMElement("userName", null);
            userName.setText(dataBaseConfiguration.getUserName());
            config.addChild(userName);
            OMElement password = factoty.createOMElement("password", null);
            password.setText(dataBaseConfiguration.getPassWord());
            config.addChild(password);
            OMElement driverName = factoty.createOMElement("driverName", null);
            driverName.setText(dataBaseConfiguration.getDriverName());
            config.addChild(driverName);
            config.addAttribute("name", dataBaseConfiguration.getConfigName(), null);
            root.addChild(config);
        }
        return root;
    }

    public DataBaseConfiguration getDeafultDataBaseConfiguration() {
        return deafultDataBaseConfiguration;
    }

    public void setDeafultDataBaseConfiguration(DataBaseConfiguration deafultDataBaseConfiguration) {
        this.deafultDataBaseConfiguration = deafultDataBaseConfiguration;
    }

    public Map getDbconfgigMap() {
        return dbconfgigMap;
    }
}
